<?php
#
# Copyright (C) 2007-2014 Hypertable, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# - Neither Hypertable, Inc. nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

if (!isset($GLOBALS['THRIFT_ROOT']))
    $GLOBALS['THRIFT_ROOT'] = getenv('PHPTHRIFT_ROOT');

require_once $GLOBALS['THRIFT_ROOT'].'/ThriftClient.php';

date_default_timezone_set("America/Los_Angeles");

function print_cell( &$cell ) {
  echo "{Cell: {Key: row=" . $cell->key->row . "' column_family='" .
    $cell->key->column_family . "' column_qualifier=";
  if (!empty($cell->key->column_qualifier))
    echo "'" . $cell->key->column_qualifier . "'";
  else
    echo "null";
  echo " ' flag=" . $cell->key->flag . "} value=";
  if (!empty($cell->value))
    echo "'" . $cell->value . "'";
  else
    echo "null";
  echo "\n";
}

function print_cell_as_array( &$cell_as_array ) {
  echo "{CellAsArray: {Key: row=" . $cell_as_array[0] . "' column_family='" .
    $cell_as_array[1] . "' column_qualifier=";
  if (!empty($cell_as_array[2]))
    echo "'" . $cell_as_array[2] . "'";
  else
    echo "null";
  echo " } value=";
  if (!empty($cell_as_array[3]))
    echo "'" . $cell_as_array[3] . "'";
  else
    echo "null";
  echo "\n";
}

function test_basic( &$client ) {

  echo "[basic]\n";

  try {

    if (!$client->namespace_exists("test"))
      $client->namespace_create("test");

    $ns = $client->namespace_open("test");

    $if_exists = True;
    $client->table_drop($ns, "Fruits", $if_exists);

    $cf = new \Hypertable_ThriftGen\ColumnFamilySpec(array('name'=> 'genus'));
    $cf_specs["genus"] = $cf;
    $cf = new \Hypertable_ThriftGen\ColumnFamilySpec(array('name'=> 'description'));
    $cf_specs["description"] = $cf;
    $cf = new \Hypertable_ThriftGen\ColumnFamilySpec(array('name'=> 'tag'));
    $cf_specs["tag"] = $cf;

    $schema = new \Hypertable_ThriftGen\Schema(array('column_families'=> $cf_specs));

    $client->table_create($ns, "Fruits", $schema);

    $client->namespace_create("/test/sub");

    $listing = $client->namespace_get_listing($ns);

    foreach ($listing as &$entry) {
      echo "$entry->name";
      if ($entry->is_namespace)
        echo "\t(dir)";
      echo "\n";
    }

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }
}

function test_convenience( &$client ) {

  echo "[set cells]\n";

  try {
    $ns = $client->namespace_open("test");

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'apple', 'column_family'=> 'genus'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> 'Malus'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'apple', 'column_family'=> 'description'));
    $value = "The apple is the pomaceous fruit of the apple tree.";
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> $value));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'apple', 'column_family'=> 'tag',
                                               'column_qualifier' => 'cruncy'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key));
    $cells[] = $cell;

    $client->set_cells($ns, "Fruits", $cells);

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[get cells]\n";

  try {
    $ns = $client->namespace_open("test");

    $columns = array("description");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('columns'=> $columns));

    $cells = $client->get_cells($ns, "Fruits", $ss);

    foreach ($cells as &$cell) {
      print_cell($cell);
    }

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[set cells as arrays]\n";

  try {
    $ns = $client->namespace_open("test");

    $cell_as_array = array("orange", "genus", "", "Citrus");
    $cells_as_arrays[] = $cell_as_array;

    $value = "The orange (specifically, the sweet orange) is the fruit of the" .
      "citrus species Citrus × sinensis in the family Rutaceae.";
    $cell_as_array = array("orange", "description", "", $value);
    $cells_as_arrays[] = $cell_as_array;

    $cell_as_array = array("orange", "tag", "juicy", "");
    $cells_as_arrays[] = $cell_as_array;

    $client->set_cells_as_arrays($ns, "Fruits", $cells_as_arrays);

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[get cells as arrays]\n";

  try {
    $ns = $client->namespace_open("test");

    $columns = array("description");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('columns'=> $columns));

    $cells_as_arrays = $client->get_cells_as_arrays($ns, "Fruits", $ss);

    foreach ($cells_as_arrays as &$cell_as_array) {
      print_cell_as_array($cell_as_array);
    }

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }
}

function test_create_table( &$client ) {

  echo "[create table]\n";

  try {
    $ns = $client->namespace_open("test");

    $ag_specs = array();
    $cf_specs = array();

    // Table defaults
    $ag_defaults = new \Hypertable_ThriftGen\AccessGroupOptions(array('blocksize'=> 65536));
    $cf_defaults = new \Hypertable_ThriftGen\ColumnFamilyOptions(array('max_versions'=> 1));

    // Access group "ag_normal"
    {
      $ag_spec = new \Hypertable_ThriftGen\AccessGroupSpec(array('name'=> 'ag_normal'));
      $cf_options = new \Hypertable_ThriftGen\ColumnFamilyOptions(array('max_versions'=> 2));
      $ag_spec->defaults = $cf_options;
      $ag_specs['ag_normal'] = $ag_spec;
    }

    // Column "a"
    {
      $args = array('name'=> 'a', 'access_group'=> 'ag_normal',
                    'value_index'=> True, 'qualifier_index'=> True);
      $cf_spec = new \Hypertable_ThriftGen\ColumnFamilySpec($args);
      $cf_specs['a'] = $cf_spec;
    }

    // Column "b"
    {
      $cf_options = new \Hypertable_ThriftGen\ColumnFamilyOptions(array('max_versions'=> 3));
      $args = array('name'=> 'b', 'access_group'=> 'ag_normal',
                    'options'=> $cf_options);
      $cf_spec = new \Hypertable_ThriftGen\ColumnFamilySpec($args);
      $cf_specs['b'] = $cf_spec;
    }

    // Access group "ag_fast"
    {
      $args = array('in_memory'=> True, 'blocksize'=> 131072);
      $ag_options = new \Hypertable_ThriftGen\AccessGroupOptions($args);
      $args = array('name'=> 'ag_fast', 'options'=> $ag_options);
      $ag_spec = new \Hypertable_ThriftGen\AccessGroupSpec($args);
      $ag_specs['ag_fast'] = $ag_spec;
    }

    // Column "c"
    {
      $args = array('name'=> 'c', 'access_group'=> 'ag_fast');
      $cf_spec = new \Hypertable_ThriftGen\ColumnFamilySpec($args);
      $cf_specs['c'] = $cf_spec;
    }

    // Access group "ag_secure"
    {
      $args = array('replication'=> 5);
      $ag_options = new \Hypertable_ThriftGen\AccessGroupOptions($args);
      $args = array('name'=> 'ag_secure', 'options'=> $ag_options);
      $ag_spec = new \Hypertable_ThriftGen\AccessGroupSpec($args);
      $ag_specs['ag_secure'] = $ag_spec;
    }

    // Column "d"
    {
      $args = array('name'=> 'd', 'access_group'=> 'ag_secure');
      $cf_spec = new \Hypertable_ThriftGen\ColumnFamilySpec($args);
      $cf_specs['d'] = $cf_spec;
    }

    // Access group "ag_counter"
    {
      $args = array('counter'=> True, 'max_versions'=> 0);
      $cf_options = new \Hypertable_ThriftGen\ColumnFamilyOptions($args);
      $args = array('name'=> 'ag_counter', 'defaults'=> $cf_options);
      $ag_spec = new \Hypertable_ThriftGen\AccessGroupSpec($args);
      $ag_specs['ag_counter'] = $ag_spec;
    }

    // Column "e"
    {
      $args = array('name'=> 'e', 'access_group'=> 'ag_counter');
      $cf_spec = new \Hypertable_ThriftGen\ColumnFamilySpec($args);
      $cf_specs['e'] = $cf_spec;
    }

    // Column "f"
    {
      $cf_options = new \Hypertable_ThriftGen\ColumnFamilyOptions(array('counter'=> False));
      $args = array('name'=> 'f', 'access_group'=> 'ag_counter',
                    'options'=> $cf_options);
      $cf_spec = new \Hypertable_ThriftGen\ColumnFamilySpec($args);
      $cf_specs['f'] = $cf_spec;
    }

    $args = array('access_group_defaults'=> $ag_defaults,
                  'column_family_defaults'=> $cf_defaults,
                  'access_groups'=> $ag_specs, 'column_families'=>$cf_specs);
    $schema = new \Hypertable_ThriftGen\Schema($args);

    $client->table_create($ns, "TestTable", $schema);

    $result = $client->hql_query($ns, "show create table TestTable");

    if (!empty($result->results))
      echo $result->results[0];

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }
}

function test_alter_table( &$client ) {

  echo "[alter table]\n";

  try {
    $ns = $client->namespace_open("test");

    $schema = $client->get_schema($ns, "TestTable");

    // Rename column "b" to "z"
    {
      $cf_spec = $schema->column_families["b"];
      unset($schema->column_families["b"]);
      $cf_spec->name = "z";
      $schema->column_families["z"] = $cf_spec;
    }

    // Add column "g"
    {
      $args = array('name'=> 'g', 'access_group'=> 'ag_counter');
      $cf_spec = new \Hypertable_ThriftGen\ColumnFamilySpec($args);
      $schema->column_families["g"] = $cf_spec;
    }

    $client->table_alter($ns, "TestTable", $schema);

    $result = $client->hql_query($ns, "show create table TestTable");

    if (!empty($result->results))
      echo $result->results[0];

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }
}

function test_mutator( &$client ) {

  echo "[mutator]\n";

  try {
    $ns = $client->namespace_open("test");

    $mutator = $client->mutator_open($ns, "Fruits", 0, 0);

    // Auto-assign timestamps

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'lemon', 'column_family'=> 'genus'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> 'Citrus'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'lemon', 'column_family'=> 'tag',
                                               'column_qualifier'=> 'bitter'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'lemon', 'column_family'=> 'description'));
    $value = "The lemon (Citrus × limon) is a small evergreen tree native to Asia.";
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> $value));
    $cells[] = $cell;

    $client->mutator_set_cells($mutator, $cells);
    $client->mutator_flush($mutator);

    // Supply timestamps

    $cells = array();

    $ts = strtotime("2014-06-06 16:27:15") * 1000000000;
    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'mango', 'column_family'=> 'genus',
                                               'timestamp'=> $ts));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> 'Mangifera'));
    $cells[] = $cell;

    $ts = strtotime("2014-06-06 16:27:15") * 1000000000;
    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'mango', 'column_family'=> 'tag',
                                               'column_qualifier'=> 'sweet',
                                               'timestamp'=> $ts));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key));
    $cells[] = $cell;

    $ts = strtotime("2014-06-06 16:27:15") * 1000000000;
    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'mango', 'column_family'=> 'description',
                                               'timestamp'=> $ts));
    $value = "Mango is one of the delicious seasonal fruits grown in the tropics.";
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> $value));
    $cells[] = $cell;

    $ts = strtotime("2014-06-06 16:27:16") * 1000000000;
    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'mango', 'column_family'=> 'description',
                                               'timestamp'=> $ts));
    $value = "The mango is a juicy stone fruit belonging to the genus " .
      "Mangifera, consisting of numerous tropical fruiting trees, that are " .
      "cultivated mostly for edible fruits.";
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> $value));
    $cells[] = $cell;

    $client->mutator_set_cells($mutator, $cells);
    $client->mutator_flush($mutator);

    // Delete cells

    $cells = array();

    $args = array('row'=> 'apple',
                  'flag'=> \Hypertable_ThriftGen\KeyFlag::DELETE_ROW);
    $key = new \Hypertable_ThriftGen\Key($args);
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key));
    $cells[] = $cell;

    $ts = strtotime("2014-06-06 16:27:15") * 1000000000;
    $args = array('row'=> 'mango', 'column_family'=> 'description',
                  'timestamp'=> $ts,
                  'flag'=> \Hypertable_ThriftGen\KeyFlag::DELETE_CELL);
    $key = new \Hypertable_ThriftGen\Key($args);
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key));
    $cells[] = $cell;

    $client->mutator_set_cells($mutator, $cells);
    $client->mutator_flush($mutator);
    $client->mutator_close($mutator);

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }
}

function test_scanner( &$client ) {

  echo "[scanner full table scan]\n";

  try {
    $ns = $client->namespace_open("test");

    $ss = new \Hypertable_ThriftGen\ScanSpec();

    $scanner = $client->scanner_open($ns, "Fruits", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[scanner restricted scan]\n";

  try {
    $ns = $client->namespace_open("test");

    // Row range [cherry..orange)
    $args = array('start_row'=> 'lemon', 'start_inclusive'=> True,
                  'end_row'=> 'orange', 'end_inclusive'=> False);
    $ri = new \Hypertable_ThriftGen\RowInterval($args);
    $row_intervals = array($ri);

    // Columns 'genus', 'tag:bitter', and 'tag:sweet'
    $columns = array('genus', 'tag:bitter', 'tag:sweet');

    $args = array('row_intervals'=> $row_intervals, 'columns'=> $columns,
                  'versions' => 1);
    $ss = new \Hypertable_ThriftGen\ScanSpec($args);

    $scanner = $client->scanner_open($ns, "Fruits", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }
}

function test_hql( &$client ) {

  echo "[HQL get listing]\n";

  try {
    $ns = $client->namespace_open("test");

    $result = $client->hql_query($ns, "GET LISTING");

    foreach ($result->results as $line)
      echo "$line\n";

    $result = $client->hql_query($ns, "SELECT * FROM Fruits WHERE ROW = 'lemon'");

    foreach ($result->cells as $cell)
      print_cell($cell);

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[HQL query cells as arrays]\n";

  try {
    $ns = $client->namespace_open("test");

    $command = "SELECT * FROM Fruits WHERE ROW = 'lemon'";
    $result = $client->hql_query_as_arrays($ns, $command);

    foreach ($result->cells as &$cell_as_array) {
      print_cell_as_array($cell_as_array);
    }

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[HQL exec - mutator]\n";

  try {
    $ns = $client->namespace_open("test");

    $command = "INSERT INTO Fruits VALUES ('strawberry', 'genus', 'Fragaria')," .
      " ('strawberry', 'tag:fibrous', ''), ('strawberry', 'description', 'The" .
      " garden strawberry is a widely grown hybrid species of the genus " .
      "Fragaria')";
    $result = $client->hql_exec($ns, $command, True, False);

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'pineapple', 'column_family'=> 'genus'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> 'Ananas'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'pineapple', 'column_family'=> 'tag',
                                               'column_qualifier'=> 'acidic'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'pineapple', 'column_family'=> 'description'));
    $value = "The pineapple (Ananas comosus) is a tropical plant with edible " .
      "multiple fruit consisting of coalesced berries.";
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> $value));
    $cells[] = $cell;

    $client->mutator_set_cells($result->mutator, $cells);
    $client->mutator_flush($result->mutator);
    $client->mutator_close($result->mutator);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[HQL exec - scanner]\n";

  try {
    $ns = $client->namespace_open("test");

    $result = $client->hql_exec($ns, "SELECT * from Fruits", False, True);

    do {
      $cells = $client->scanner_get_cells($result->scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($result->scanner);

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }
}

function test_secondary_indices( &$client ) {

  echo "[secondary indices]\n";

  try {
    $ns = $client->namespace_open("test");

    $command = "CREATE TABLE products (title, section, info, category," .
      "INDEX section, INDEX info, QUALIFIER INDEX info," .
      "QUALIFIER INDEX category)";
    $result = $client->hql_query($ns, $command);

    $command = "LOAD DATA INFILE 'indices_test_products.tsv' INTO TABLE products";
    $result = $client->hql_query($ns, $command);

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[secondary index - SELECT title FROM products WHERE section = 'books']\n";

  try {
    $ns = $client->namespace_open("test");

    $column_predicates = array();
    $args = array('column_family'=> 'section',
                  'operation'=> \Hypertable_ThriftGen\ColumnPredicateOperation::EXACT_MATCH,
                  'value'=> 'books');
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $columns = array("title");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('column_predicates'=> $column_predicates,
                                                   'columns'=> $columns));

    $scanner = $client->scanner_open($ns, "products", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[secondary index - SELECT title FROM products WHERE info:actor = 'Jack Nicholson']\n";

  try {
    $ns = $client->namespace_open("test");

    $column_predicates = array();
    $operation = \Hypertable_ThriftGen\ColumnPredicateOperation::EXACT_MATCH |
      \Hypertable_ThriftGen\ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
    $args = array('column_family'=> 'info', 'column_qualifier'=> 'actor',
                  'operation'=> $operation, 'value'=> 'Jack Nicholson');
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $columns = array("title");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('column_predicates'=> $column_predicates,
                                                   'columns'=> $columns));

    $scanner = $client->scanner_open($ns, "products", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[secondary index - SELECT title,info:publisher FROM products WHERE info:publisher =^ 'Addison-Wesley']\n";

  try {
    $ns = $client->namespace_open("test");

    $column_predicates = array();
    $operation = \Hypertable_ThriftGen\ColumnPredicateOperation::PREFIX_MATCH |
      \Hypertable_ThriftGen\ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
    $args = array('column_family'=> 'info', 'column_qualifier'=> 'publisher',
                  'operation'=> $operation, 'value'=> 'Addison-Wesley');
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $columns = array("title", "info:publisher");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('column_predicates'=> $column_predicates,
                                                   'columns'=> $columns));

    $scanner = $client->scanner_open($ns, "products", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[secondary index - SELECT title,info:publisher FROM products WHERE info:publisher =~ /^Addison-Wesley/]\n";

  try {
    $ns = $client->namespace_open("test");

    $column_predicates = array();
    $operation = \Hypertable_ThriftGen\ColumnPredicateOperation::REGEX_MATCH |
      \Hypertable_ThriftGen\ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
    $args = array('column_family'=> 'info', 'column_qualifier'=> 'publisher',
                  'operation'=> $operation, 'value'=> '^Addison-Wesley');
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $columns = array("title", "info:publisher");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('column_predicates'=> $column_predicates,
                                                   'columns'=> $columns));

    $scanner = $client->scanner_open($ns, "products", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[secondary index - SELECT title FROM products WHERE Exists(info:studio)]\n";

  try {
    $ns = $client->namespace_open("test");

    $column_predicates = array();
    $operation = \Hypertable_ThriftGen\ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
    $args = array('column_family'=> 'info', 'column_qualifier'=> 'studio',
                  'operation'=> $operation);
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $columns = array("title");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('column_predicates'=> $column_predicates,
                                                   'columns'=> $columns));

    $scanner = $client->scanner_open($ns, "products", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[secondary index - SELECT title FROM products WHERE Exists(category:/^\\/Movies/)]\n";

  try {
    $ns = $client->namespace_open("test");

    $column_predicates = array();
    $operation = \Hypertable_ThriftGen\ColumnPredicateOperation::QUALIFIER_REGEX_MATCH;
    $args = array('column_family'=> 'category', 'column_qualifier'=> '^/Movies',
                  'operation'=> $operation);
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $columns = array("title");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('column_predicates'=> $column_predicates,
                                                   'columns'=> $columns));

    $scanner = $client->scanner_open($ns, "products", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[secondary index - SELECT title FROM products " .
    "WHERE info:author =~ /^Stephen P/ OR info:publisher =^ 'Anchor']\n";

  try {
    $ns = $client->namespace_open("test");

    $column_predicates = array();
    $operation = \Hypertable_ThriftGen\ColumnPredicateOperation::REGEX_MATCH |
      \Hypertable_ThriftGen\ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
    $args = array('column_family'=> 'info', 'column_qualifier'=> 'author',
                  'operation'=> $operation, 'value'=> '^Stephen P');
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $operation = \Hypertable_ThriftGen\ColumnPredicateOperation::PREFIX_MATCH |
      \Hypertable_ThriftGen\ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
    $args = array('column_family'=> 'info', 'column_qualifier'=> 'publisher',
                  'operation'=> $operation, 'value'=> 'Anchor');
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $columns = array("title");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('column_predicates'=> $column_predicates,
                                                   'columns'=> $columns));

    $scanner = $client->scanner_open($ns, "products", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[secondary index - SELECT title FROM products WHERE " .
    "info:author =~ /^Stephen [PK]/ AND info:publisher =^ 'Anchor']\n";

  try {
    $ns = $client->namespace_open("test");

    $column_predicates = array();
    $operation = \Hypertable_ThriftGen\ColumnPredicateOperation::REGEX_MATCH |
      \Hypertable_ThriftGen\ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
    $args = array('column_family'=> 'info', 'column_qualifier'=> 'author',
                  'operation'=> $operation, 'value'=> '^Stephen [PK]');
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $operation = \Hypertable_ThriftGen\ColumnPredicateOperation::PREFIX_MATCH |
      \Hypertable_ThriftGen\ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
    $args = array('column_family'=> 'info', 'column_qualifier'=> 'publisher',
                  'operation'=> $operation, 'value'=> 'Anchor');
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $columns = array("title");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('column_predicates'=> $column_predicates,
                                                   'columns'=> $columns,
                                                   'and_column_predicates'=> True));

    $scanner = $client->scanner_open($ns, "products", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[secondary index - SELECT title FROM products WHERE ROW > 'B00002VWE0'" .
    " AND info:actor = 'Jack Nicholson']\n";

  try {
    $ns = $client->namespace_open("test");

    // ROW > 'B00002VWE0'
    $args = array('start_row'=> 'B00002VWE0', 'start_inclusive'=> False);
    $ri = new \Hypertable_ThriftGen\RowInterval($args);
    $row_intervals = array($ri);

    // info:actor = 'Jack Nicholson'
    $column_predicates = array();
    $operation = \Hypertable_ThriftGen\ColumnPredicateOperation::EXACT_MATCH |
      \Hypertable_ThriftGen\ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
    $args = array('column_family'=> 'info', 'column_qualifier'=> 'actor',
                  'operation'=> $operation, 'value'=> 'Jack Nicholson');
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $columns = array("title");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('row_intervals'=> $row_intervals,
                                                   'column_predicates'=> $column_predicates,
                                                   'columns'=> $columns,
                                                   'and_column_predicates'=> True));

    $scanner = $client->scanner_open($ns, "products", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[secondary index - SELECT title FROM products " .
    "WHERE ROW =^ 'B' AND info:actor = 'Jack Nicholson']\n";

  try {
    $ns = $client->namespace_open("test");

    // ROW =^ 'B'
    $args = array('start_row'=> 'B', 'start_inclusive'=> True,
                  'end_row'=> 'C', 'end_inclusive'=> False);
    $ri = new \Hypertable_ThriftGen\RowInterval($args);
    $row_intervals = array($ri);

    // info:actor = 'Jack Nicholson'
    $column_predicates = array();
    $operation = \Hypertable_ThriftGen\ColumnPredicateOperation::EXACT_MATCH |
      \Hypertable_ThriftGen\ColumnPredicateOperation::QUALIFIER_EXACT_MATCH;
    $args = array('column_family'=> 'info', 'column_qualifier'=> 'actor',
                  'operation'=> $operation, 'value'=> 'Jack Nicholson');
    $column_predicate = new \Hypertable_ThriftGen\ColumnPredicate($args);
    $column_predicates[] = $column_predicate;

    $columns = array("title");

    $ss = new \Hypertable_ThriftGen\ScanSpec(array('row_intervals'=> $row_intervals,
                                                   'column_predicates'=> $column_predicates,
                                                   'columns'=> $columns,
                                                   'and_column_predicates'=> True));

    $scanner = $client->scanner_open($ns, "products", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }
}

function test_async( &$client ) {

  echo "[async]\n";

  try {
    $ns = $client->namespace_open("test");

    $command = "CREATE TABLE Profile (info, last_access MAX_VERSIONS 1)";
    $result = $client->hql_query($ns, $command);

    $command = "CREATE TABLE Session (user_id, page_hit)";
    $result = $client->hql_query($ns, $command);

    $command = "INSERT INTO Profile VALUES ('1', 'info:name', 'Joe'), ('2', 'info:name', 'Sue')";
    $result = $client->hql_query($ns, $command);

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[async mutator]\n";

  try {
    $ns = $client->namespace_open("test");

    $ff = $client->future_open(0);
    $profile_mutator = $client->async_mutator_open($ns, "Profile", $ff, 0);
    $session_mutator = $client->async_mutator_open($ns, "Session", $ff, 0);

    $cells = array();

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '1', 'column_family'=> 'last_access'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '2014-06-13 16:06:09'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '2', 'column_family'=> 'last_access'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '2014-06-13 16:06:10'));
    $cells[] = $cell;

    $client->async_mutator_set_cells($profile_mutator, $cells);

    $cells = array();

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '0001-200238',
                                               'column_family'=> 'user_id'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '0001-200238',
                                               'column_family'=> 'page_hit'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '/index.html'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '0002-383049',
                                               'column_family'=> 'user_id'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '0002-383049',
                                               'column_family'=> 'page_hit'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '/foo/bar.html'));
    $cells[] = $cell;

    $client->async_mutator_set_cells($session_mutator, $cells);

    $client->async_mutator_flush($profile_mutator);
    $client->async_mutator_flush($session_mutator);

    $result_count = 0;
    while (True) {
      $result = $client->future_get_result($ff, 0);
      if ($result->is_empty)
        break;
      $result_count++;
      if ($result->is_error) {
        echo "Async mutator error:  " . $result->error_msg . "\n";
        exit(1);
      }
      if ($result->id == $profile_mutator)
        echo "Result is from Profile mutation\n";
      else if ($result->id == $session_mutator)
        echo "Result is from Session mutation\n";
    }

    echo "result count = " . $result_count . "\n";

    $client->async_mutator_close($profile_mutator);
    $client->async_mutator_close($session_mutator);
    $client->future_close($ff);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[async scanner]\n";

  try {
    $ns = $client->namespace_open("test");

    $ff = $client->future_open(0);

    // SELECT * from Profile WHERE ROW = "1"
    $args = array('start_row'=> '1', 'start_inclusive'=> True,
                  'end_row'=> '1', 'end_inclusive'=> True);
    $ri = new \Hypertable_ThriftGen\RowInterval($args);
    $row_intervals = array($ri);
    $ss = new \Hypertable_ThriftGen\ScanSpec(array('row_intervals'=> $row_intervals));
    $profile_scanner = $client->async_scanner_open($ns, "Profile", $ff, $ss);

    // SELECT * from Session WHERE ROW = "0001-200238"
    $args = array('start_row'=> '0001-200238', 'start_inclusive'=> True,
                  'end_row'=> '0001-200238', 'end_inclusive'=> True);
    $ri = new \Hypertable_ThriftGen\RowInterval($args);
    $row_intervals = array($ri);
    $ss = new \Hypertable_ThriftGen\ScanSpec(array('row_intervals'=> $row_intervals));
    $session_scanner = $client->async_scanner_open($ns, "Session", $ff, $ss);

    while (True) {

      $result = $client->future_get_result($ff, 0);

      if ($result->is_empty)
        break;

      if ($result->is_error) {
        echo "Async scanner error:  " . $result->error_msg . "\n";
        exit(1);
      }

      assert($result->is_scan);
      assert($result->id == $profile_scanner || $result->id == $session_scanner);

      if ($result->id == $profile_scanner)
        echo "Result is from Profile scan\n";
      else if ($result->id == $session_scanner)
        echo "Result is from Session scan\n";

      foreach ($result->cells as &$cell) {
        print_cell($cell);
      }
    }

    $client->async_scanner_close($profile_scanner);
    $client->async_scanner_close($session_scanner);
    $client->future_close($ff);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[async scanner - result as arrays]\n";

  try {
    $ns = $client->namespace_open("test");

    $ff = $client->future_open(0);

    // SELECT * from Profile WHERE ROW = "1"
    $args = array('start_row'=> '1', 'start_inclusive'=> True,
                  'end_row'=> '1', 'end_inclusive'=> True);
    $ri = new \Hypertable_ThriftGen\RowInterval($args);
    $row_intervals = array($ri);
    $ss = new \Hypertable_ThriftGen\ScanSpec(array('row_intervals'=> $row_intervals));
    $profile_scanner = $client->async_scanner_open($ns, "Profile", $ff, $ss);

    // SELECT * from Session WHERE ROW = "0001-200238"
    $args = array('start_row'=> '0001-200238', 'start_inclusive'=> True,
                  'end_row'=> '0001-200238', 'end_inclusive'=> True);
    $ri = new \Hypertable_ThriftGen\RowInterval($args);
    $row_intervals = array($ri);
    $ss = new \Hypertable_ThriftGen\ScanSpec(array('row_intervals'=> $row_intervals));
    $session_scanner = $client->async_scanner_open($ns, "Session", $ff, $ss);

    while (True) {

      $result_as_arrays = $client->future_get_result_as_arrays($ff, 0);

      if ($result_as_arrays->is_empty)
        break;

      if ($result_as_arrays->is_error) {
        echo "Async scanner error:  " . $result_as_arrays->error_msg . "\n";
        exit(1);
      }

      assert($result_as_arrays->is_scan);
      assert($result_as_arrays->id == $profile_scanner ||
             $result_as_arrays->id == $session_scanner);

      if ($result_as_arrays->id == $profile_scanner)
        echo "Result is from Profile scan\n";
      else if ($result_as_arrays->id == $session_scanner)
        echo "Result is from Session scan\n";

      foreach ($result_as_arrays->cells as &$cell_as_array) {
        print_cell_as_array($cell_as_array);
      }
    }

    $client->async_scanner_close($profile_scanner);
    $client->async_scanner_close($session_scanner);
    $client->future_close($ff);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }
}


function test_atomic_counter( &$client ) {

  echo "[atomic counter]\n";

  try {
    $ns = $client->namespace_open("test");

    $command = "CREATE TABLE Hits (count COUNTER)";
    $result = $client->hql_query($ns, $command);

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[atomic counter - increment]\n";

  try {
    $ns = $client->namespace_open("test");

    $mutator = $client->mutator_open($ns, "Hits", 0, 0);

    $cells = array();

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/index.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:18'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '1'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/index.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:18'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '1'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/foo/bar.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:18'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '1'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/foo/bar.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:18'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '1'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/foo/bar.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:18'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '1'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/index.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:19'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '1'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/index.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:19'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '1'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/index.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:19'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '1'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/index.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:19'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '1'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/foo/bar.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:19'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '1'));
    $cells[] = $cell;

    $client->mutator_set_cells($mutator, $cells);
    $client->mutator_flush($mutator);

    $ss = new \Hypertable_ThriftGen\ScanSpec();

    $scanner = $client->scanner_open($ns, "Hits", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  echo "[atomic counter - reset and subtraction]\n";

  try {
    $ns = $client->namespace_open("test");

    $mutator = $client->mutator_open($ns, "Hits", 0, 0);

    $cells = array();

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/index.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:18'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '=0'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/index.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:18'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '7'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/foo/bar.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:18'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '-1'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/index.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:19'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '-2'));
    $cells[] = $cell;

    $key = new \Hypertable_ThriftGen\Key(array('row'=> '/foo/bar.html',
                                               'column_family'=> 'count',
                                               'column_qualifier'=> '2014-06-14 07:31:19'));
    $cell = new \Hypertable_ThriftGen\Cell(array('key' => $key, 'value'=> '=19'));
    $cells[] = $cell;

    $client->mutator_set_cells($mutator, $cells);
    $client->mutator_flush($mutator);

    $ss = new \Hypertable_ThriftGen\ScanSpec();

    $scanner = $client->scanner_open($ns, "Hits", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);
    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }
}

function test_unique( &$client ) {

  echo "[unique]\n";

  try {
    $ns = $client->namespace_open("test");

    $command = "CREATE TABLE User (info, id TIME_ORDER desc MAX_VERSIONS 1)";

    $result = $client->hql_query($ns, $command);

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }

  try {
    $ns = $client->namespace_open("test");

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'joe1987',
                                               'column_family'=> 'id'));
    $ret = $client->create_cell_unique($ns, "User", $key, "");

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'mary.bellweather',
                                               'column_family'=> 'id'));
    $ret = $client->create_cell_unique($ns, "User", $key, "");

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }


  try {
    $ns = $client->namespace_open("test");

    $key = new \Hypertable_ThriftGen\Key(array('row'=> 'joe1987',
                                               'column_family'=> 'id'));
    $ret = $client->create_cell_unique($ns, "User", $key, "");

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    if ($e->code == 48) { // ALREADY_EXISTS
      echo "User name '" . $key->row . "' is already taken\n";
      $client->namespace_close($ns);
    }
    else {
      echo "error: $e->message\n";
      exit(1);
    }
  }

  try {

    $ns = $client->namespace_open("test");

    $ss = new \Hypertable_ThriftGen\ScanSpec();

    $scanner = $client->scanner_open($ns, "User", $ss);

    do {
      $cells = $client->scanner_get_cells($scanner);
      foreach ($cells as &$cell) {
        print_cell($cell);
      }
    } while (!empty($cells));

    $client->scanner_close($scanner);

    $client->namespace_close($ns);
  }
  catch (\Hypertable_ThriftGen\ClientException $e) {
    echo "error: $e->message\n";
    exit(1);
  }
}

try {
  $client = new Hypertable_ThriftClient("localhost", 15867);
}
catch (\Hypertable_ThriftGen\ClientException $e) {
  echo "error: $e->message\n";
  exit(1);
}

test_basic($client);
test_convenience($client);
test_create_table($client);
test_alter_table($client);
test_mutator($client);
test_scanner($client);
test_hql($client);
test_secondary_indices($client);
test_async($client);
test_atomic_counter($client);
test_unique($client);
